from __future__ import annotations

from dataclasses import dataclass

from aau_label.model import AAULabel
from numpy.typing import NDArray


@dataclass
class Size:
    width: float
    height: float


@dataclass
class ImgTensor:
    scale: Size
    data: NDArray


@dataclass
class ClassifierResult:
    name: str
    score: float


@dataclass
class DetectorResult(ClassifierResult, AAULabel): ...



import numpy
from numpy import float32, float64, int32, int64
from numpy.typing import NDArray


def compute_iou(box: NDArray[int32], boxes: NDArray[int32]) -> NDArray[float64]:
    # Compute xmin, ymin, xmax, ymax for both boxes
    xmin = numpy.minimum(box[0], boxes[:, 0])
    ymin = numpy.minimum(box[1], boxes[:, 1])
    xmax = numpy.maximum(box[0] + box[2], boxes[:, 0] + boxes[:, 2])
    ymax = numpy.maximum(box[1] + box[3], boxes[:, 1] + boxes[:, 3])

    # Compute intersection area
    intersection_area = numpy.maximum(0, xmax - xmin) * numpy.maximum(0, ymax - ymin)

    # Compute union area
    box_area = box[2] * box[3]
    boxes_area = boxes[:, 2] * boxes[:, 3]
    union_area = box_area + boxes_area - intersection_area

    # Compute IoU
    iou = intersection_area / union_area

    return iou


def nms(
    boxes: NDArray[int32], scores: NDArray[float32], iou_threshold: float
) -> list[int64]:
    # Sort by score
    sorted_indices = numpy.argsort(scores)[::-1]

    keep_boxes = []
    while sorted_indices.size > 0:
        # Pick the last box
        box_id = sorted_indices[0]
        keep_boxes.append(box_id)

        # Compute IoU of the picked box with the rest
        ious = compute_iou(boxes[box_id, :], boxes[sorted_indices[1:], :])

        # Remove boxes with IoU over the threshold
        keep_indices = numpy.where(ious < iou_threshold)[0]

        # print(keep_indices.shape, sorted_indices.shape)
        sorted_indices = sorted_indices[keep_indices + 1]

    return keep_boxes



import ast
import logging
from typing import Generic, Sequence, TypeVar, cast

import numpy
from numpy.typing import NDArray

from .. import utils
from ..model import DetectorResult, ImgTensor
from ..protocols import InferenceSessionProtocol, ToTensorStrategyProtocol

T = TypeVar("T")
logger = logging.getLogger(__name__)


class Detector(Generic[T]):
    def __init__(
        self,
        session: InferenceSessionProtocol,
        to_tensor_strategy: ToTensorStrategyProtocol[T],
        conf_threshold: float = 0.25,
        iou_threshold: float = 0.7,
    ) -> None:
        self.__session = session
        self.__to_tensor_strategy = to_tensor_strategy
        self.__conf_threshold = conf_threshold
        self.__iou_threshold = iou_threshold

        meta = self.__session.get_modelmeta()
        self.__names: dict[int, str] = ast.literal_eval(
            meta.custom_metadata_map["names"]
        )

    @property
    def conf_threshold(self) -> float:
        return self.__conf_threshold

    @property
    def iou_threshold(self) -> float:
        return self.__iou_threshold

    @property
    def shape(self) -> tuple[int, int]:
        return self.__session.get_inputs()[0].shape[2:]

    @property
    def names(self) -> dict[int, str]:
        return self.__names

    def __result_handler(
        self, results: NDArray, tensor: ImgTensor
    ) -> Sequence[DetectorResult]:
        predictions = numpy.squeeze(results[0]).T

        scores = numpy.max(predictions[:, 4:], axis=1)
        keep = scores > self.__conf_threshold
        predictions = predictions[keep, :]
        scores = scores[keep]
        class_ids = numpy.argmax(predictions[:, 4:], axis=1)

        boxes = predictions[:, :4]
        # Make x0, y0 left upper corner instead of box center
        boxes[:, 0:2] -= boxes[:, 2:4] / 2
        boxes /= numpy.array(
            [
                tensor.scale.width,
                tensor.scale.height,
                tensor.scale.width,
                tensor.scale.height,
            ],
            dtype=numpy.float32,
        )
        boxes = boxes.astype(numpy.int32)

        keep = utils.nms(boxes, scores, self.__iou_threshold)
        rv = []
        for bbox, label, score in zip(boxes[keep], class_ids[keep], scores[keep]):
            rv.append(
                DetectorResult(
                    x=bbox[0].item(),
                    y=bbox[1].item(),
                    width=bbox[2].item(),
                    height=bbox[3].item(),
                    name=self.__names[label],
                    score=score.item(),
                )
            )

        return rv

    def run(self, img: T) -> Sequence[DetectorResult]:
        tensor = self.__to_tensor_strategy(img, *self.shape)
        results = cast(list[NDArray], self.__session.run(None, {"images": tensor.data}))
        return self.__result_handler(results[0], tensor)
